// Serialize/deserialize basic types
// Agreement: most significant byte first
//
#include "serialize.h"
#include <string.h>


void u16_to_bytes(u8 *bytes, const u16 x)
{
bytes[0] = (u8)(x>>8);
bytes[1] = (u8)(x&0xFF);
}

void u32_to_bytes(u8 *bytes, const u32 x)
{
u16_to_bytes(bytes, (u16)(x>>16));
u16_to_bytes(bytes + 2,(u16)(x & 0xFFFF));
}

void u64_to_bytes(u8 *bytes, const u64 x)
{
u32_to_bytes(bytes, (u32)(x>>32));
u32_to_bytes(bytes + 4, (u32)(x & 0xFFFFFFFF));
}

u16 bytes_to_u16(u8 const* bytes)
{
return (((u16)bytes[0])<<8) | bytes[1];
}

u32 bytes_to_u32(u8 const* bytes)
{
return (((u32)bytes_to_u16(bytes))<<16) | bytes_to_u16(bytes+2);
}

u64 bytes_to_u64(u8 const* bytes)
{
return (((u64)bytes_to_u32(bytes))<<32) | bytes_to_u32(bytes+4);
}

// Convert u32[] to byte[]

void serialize_u32(u8 *bytes, u32 const* data, const u32 length_x32)
{
for(u32 ci = 0; ci < length_x32; ci++)
 {
 u32_to_bytes(bytes + 4*ci, data[ci]);
 }
}

void deserialize_u32(u32 *data, u8 const* bytes, const u32 length_x32) // Convert u8[] to u32[]
{
for(u32 ci = 0; ci < length_x32; ci++)
 {
 data[ci] = bytes_to_u32(bytes + 4*ci);
 }
}
//--------------------------------
// Portable 
// Least significant byte first
// Returns remaining number of bytes
//
u32 fetch_dwords(u32 *data_dwords, const u32 number_of_dwords, u8 const *bytes, u32 length_bytes)
{
    
memset(data_dwords, 0, number_of_dwords*sizeof(u32));

for(u32 i = 0; i < number_of_dwords; i++)
 {
 u32 shft = 0;
 u32 d = 0;  
 while(length_bytes)
  {
  length_bytes--;
  u32 x = (u32)(*bytes++);
  d |= x << shft;
  shft += 8;
  if(shft >= 32) break;
  }
 *data_dwords++ = d;
 }

return length_bytes;
}


//=============== <END> =============
