#include <tchar.h>
#include "control.h"
#include "resource.h"
#include "main.h"
//----------------------------------------------------------------
CONTR_PROT::CONTR_PROT(NODE* node, HWND mWnd, HWND parent)
{
	mainWnd = mWnd;
	this->parent = parent;
    nd = *node;
    state = CTRL_ST_IDLE;
    msg = NULL;
    send_msg_num = 0;
    rec_msg_num = 0;
	rec_msg_flag = 0;
	first_msg_rcv = false;
    for (int i = 0; i < CONTR_MAX_VOICE_BUF; i++)
	{
        rec_data[i] = new u8[1024];
        rec_size[i] = 0;
	}
	rec_wr_inx = 0;
	rec_rd_inx = 0;
	rec_pack_num = 0;
	send_pack_num = 0;
	rec_data_buf_limit = 8;
	voice_cnt = 0;
	total_pack_num = 0;
	voice_pack_num = 0;
	voice_pack_proc = 0;
	long_msg = NULL;
	ZeroMemory(long_msg_mask,sizeof(long_msg_mask));
	long_msg_size = 0;
	InitializeCriticalSection(&cs);
	type_ticks = GetTickCount();
}
//----------------------------------------------------------------
CONTR_PROT::~CONTR_PROT()
{
    if (msg)
    {
        MESSAGE* cur;
        for (cur=msg;cur;)
        {
            MESSAGE* next = cur->next;
			delete cur->text;
            delete cur;
            cur = next;
        }
    }
	for (int i = 0; i < CONTR_MAX_VOICE_BUF; i++)
	{
		delete[] rec_data[i];
	}
	DeleteCriticalSection(&cs);
}
//----------------------------------------------------------------
static void TransferMessage(HWND hWnd, MESSAGE* message,NODE* nd)
{
	u8 buf[NETWORK_SIZE_USEFUL] = {0};
	if (message->len <= NETWORK_SIZE_TEXT)
	{
		buf[0] = CMD_SHOW_MESSAGE;
		*(u32*)&buf[1] = message->msg_num;
		CopyMemory(&buf[1+4],message->text,message->len);
		PACKET_FOR_SEND pack = {0};
		pack.buffer = buf;
		pack.buf_len = 5 + message->len;
		pack.nd = nd;
		pack.protocol = NETWORK_PROT_CONTROL;
		getNet()->send_packet(&pack);
		message->send_cnt++;
	}
	else
	{
		buf[0] = CMD_SHOW_LONG_MESSAGE;
		*(u32*)&buf[1] = message->msg_num;
		*(u16*)&buf[1+4] = NETWORK_SIZE_LONG_TEXT;
		buf[1+4+2] = (u8)((message->len+NETWORK_SIZE_LONG_TEXT-1)/NETWORK_SIZE_LONG_TEXT);
		for (u32 sent = 0; sent < message->len;)
		{
			u32 len = message->len-sent;
			if (len > NETWORK_SIZE_LONG_TEXT) len = NETWORK_SIZE_LONG_TEXT;
			buf[1+4+2+1] = (u8)(sent/NETWORK_SIZE_LONG_TEXT);
			CopyMemory(&buf[1+4+4],&message->text[sent],len);
			ZeroMemory(&buf[1+4+4]+len,NETWORK_SIZE_LONG_TEXT-len);
			PACKET_FOR_SEND pack = {0};
			pack.buffer = buf;
			pack.buf_len = 1+4+4 + NETWORK_SIZE_LONG_TEXT;
			pack.nd = nd;
			pack.protocol = NETWORK_PROT_CONTROL;
			pack.dont_count_tx=true;
			getNet()->send_packet(&pack);
			sent += len;
		}
		message->send_cnt++;
	}
}
//----------------------------------------------------------------
void CONTR_PROT::sendMessage(TCHAR* message,LONG len,LONG posMin, LONG posMax)
{
    MESSAGE* ptr;
    if (state == CTRL_ST_IDLE) {
		state = CTRL_ST_CONNECTION;
		process();
        send_msg_num = 0;
        rec_msg_num = 0;
		PostMessage(parent,WM_UPDATECHAT,0,0);
	}
    if (msg)
    {
        for (ptr = msg; ptr->next; ptr = ptr->next);
        ptr->next = new MESSAGE;
        ptr = ptr->next;
    }
    else
    {
        msg = new MESSAGE;
        ptr = msg;
    }
    ptr->msg_num = send_msg_num++;
    ptr->next = NULL;
    ptr->send_cnt = 0;
	ptr->len = len*sizeof(TCHAR);
    ptr->text = new u8[ptr->len];
	memcpy(ptr->text,message,ptr->len);
	ptr->min_pos = posMin;
	ptr->max_pos = posMax;

    TransferMessage(mainWnd,ptr,&nd);
}
//----------------------------------------------------------------
void CONTR_PROT::sendVoice(u8* data, u32 size)
{
	u8 send_voice_buf[NETWORK_SIZE_USEFUL];
	PACKET_FOR_SEND pack = {0};
	if (state != CTRL_ST_IN_VOICE || size > NETWORK_SIZE_USEFUL-1) return;
	send_voice_buf[0] = CMD_VOICE;
	CopyMemory(&send_voice_buf[1],data,size);
	pack.buffer = send_voice_buf;
	pack.buf_len = 1 + size;
	pack.nd = &nd;
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.dont_count_tx = true;
	getNet()->send_packet(&pack);
}
//----------------------------------------------------------------
void CONTR_PROT::processMessages(void)
{
    MESSAGE* ptr;
    for (ptr = msg; ptr; ptr = ptr->next)
    {
        TransferMessage(parent,ptr,&nd);
    }
}
//----------------------------------------------------------------
void CONTR_PROT::process(void)
{
	PACKET_FOR_SEND pack = {0};
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;
	NETWORK* net = getNet();

    switch(state)
    {
    case CTRL_ST_IDLE:
    default:
		voice_cnt = 0;
        break;
    case CTRL_ST_CONNECTION:
		pack.buffer = &CMD_START_CONNECTION;
		pack.buf_len = sizeof(CMD_START_CONNECTION);
		if (net) net->send_packet(&pack);
        break;
    case CTRL_ST_CONNECTED:
    case CTRL_ST_ABORT:
    case CTRL_ST_DIALING:
    case CTRL_ST_IN_VOICE:
        processMessages();
        break;
    }
	if (state == CTRL_ST_ABORT)
	{
        voice_cnt++;
        if (voice_cnt > 3)
        {
            state = CTRL_ST_IDLE;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
	}
    if (state == CTRL_ST_DIALING)
	{
        voice_cnt++;
		pack.buffer = &CMD_HANG_UP;
		pack.buf_len = sizeof(CMD_HANG_UP);
		if (net) net->send_packet(&pack);
        if (voice_cnt > 60)
        {
            state = CTRL_ST_IDLE;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
	}
    if (state == CTRL_ST_IN_VOICE)
    {
        voice_cnt++;
		pack.buffer = &CMD_IN_VOICE;
		pack.buf_len = sizeof(CMD_IN_VOICE);
		if (net) net->send_packet(&pack);
        if (voice_cnt > 20)
        {
            state = CTRL_ST_DIALING;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
    }
}

void CONTR_PROT::received(u8* packet, u32 size)
{
 	PACKET_FOR_SEND pack = {0};
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;
	NETWORK* net = getNet();

	total_pack_num++;

	if (size < 1) return;
	voice_cnt = 0;
    switch(packet[0]) // command
    {
    case CMD_START_CONNECTION:
		pack.buffer = &CMD_CONNECTED;
		pack.buf_len = sizeof(CMD_CONNECTED);
		if (net) net->send_packet(&pack);
        if (state != CTRL_ST_CONNECTED)
        {
            state = CTRL_ST_CONNECTED;
            send_msg_num = 0;
            rec_msg_num = 0;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
        break;
    case CMD_CONNECTED:
        if (state < CTRL_ST_CONNECTED)
        {
            state = CTRL_ST_CONNECTED;
			PostMessage(parent,WM_UPDATECHAT,0,0);
        }
        break;
    case CMD_SHOW_MESSAGE:
        if(size > 1+sizeof(u32))
        {
			// send responce
            u8 data[1+4] = {CMD_MESSAGE_RECEIVED,packet[1],packet[2],packet[3],packet[4]};
			pack.buffer = data;
			pack.buf_len = sizeof(data);
			if (net) net->send_packet(&pack);
			// parse packet
			u32 cur_num = *(u32*)&packet[1];
			if (cur_num > rec_msg_num)
            {
				SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[5],size-1-sizeof(u32),0,100,NULL);
				rec_msg_flag >>= cur_num - rec_msg_num;
                rec_msg_num = cur_num;
				rec_msg_flag |= 0x80000000;
				first_msg_rcv = false;
            }
			else if ((rec_msg_num && !cur_num) || ((rec_msg_num - cur_num) > 32) || (!rec_msg_num && !cur_num && !first_msg_rcv)) // session restart
			{
				SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[5],size-1-sizeof(u32),0,100,NULL);
				rec_msg_num = cur_num;
				rec_msg_flag = 0x80000000;
				first_msg_rcv = true;
			}
			else
			{
				u8 shift = rec_msg_num - cur_num;
				u32 mask = 0x80000000 >> shift;
				if (!(rec_msg_flag & mask))
				{
					SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)&packet[5],size-1-sizeof(u32),0,100,NULL);
				}
				rec_msg_flag |= mask;
			}
        }
        break;
	case CMD_SHOW_LONG_MESSAGE:
        if(size > 1+sizeof(u32)*2)
        {
			// parse packet
			u32 cur_num = *(u32*)&packet[1];
			u32 chunk_size = *(u16*)&packet[1+4];
			u8 chunk_total = packet[1+4+2];
			u8 chunk_num = packet[1+4+2+1];
			if (size >= 1+sizeof(u32)*2 + chunk_size)
			{
				if (cur_num > rec_msg_num)
				{
					if (long_msg) delete[] long_msg;
					long_msg = NULL;
					long_msg_size = 0;
					ZeroMemory(long_msg_mask,sizeof(long_msg_mask));
					rec_msg_flag >>= cur_num - rec_msg_num;
					rec_msg_num = cur_num;
				}
				else if ((rec_msg_num && !cur_num) || ((rec_msg_num - cur_num) > 32) || (!rec_msg_num && !cur_num && !first_msg_rcv)) // session restart
				{
					if (long_msg) delete[] long_msg;
					long_msg = NULL;
					long_msg_size = 0;
					ZeroMemory(long_msg_mask,sizeof(long_msg_mask));
					rec_msg_num = cur_num;
					rec_msg_flag = 0;
					first_msg_rcv = true;
				}
				if (cur_num == rec_msg_num)
				{
					if (!long_msg)
					{
						long_msg_size = chunk_size*chunk_total;
						long_msg = new u8[long_msg_size+sizeof(TCHAR)];
						if (long_msg) ZeroMemory(long_msg,long_msg_size+sizeof(TCHAR));
					}
					if (long_msg && (long_msg_size == chunk_size*chunk_total) && (chunk_num < chunk_total))
					{
						u8 chunk;
						memcpy(&long_msg[chunk_num*chunk_size],&packet[1+8],chunk_size);
						long_msg_mask[chunk_num>>3] |= (1 << (chunk_num&7));
						for (chunk = 0; chunk < chunk_total; chunk++)
						{
							if (!(long_msg_mask[chunk>>3] & (1 << (chunk&7)))) break;
						}
						if (chunk == chunk_total) // everything is received
						{
							if (!(rec_msg_flag & 0x80000000))
							{
								SendMessageTimeout(parent,WM_SHOWTEXT,(WPARAM)long_msg,long_msg_size,0,100,NULL);
							}
							rec_msg_flag |= 0x80000000;
							first_msg_rcv = false;
							delete[] long_msg;
							long_msg = NULL;
							long_msg_size = 0;
							ZeroMemory(long_msg_mask,sizeof(long_msg_mask));
							// send responce
							u8 data[1+4] = {CMD_MESSAGE_RECEIVED,packet[1],packet[2],packet[3],packet[4]};
							pack.buffer = data;
							pack.buf_len = sizeof(data);
							if (net) net->send_packet(&pack);
						}
					}
					else // wrong message
					{
						if (long_msg) delete[] long_msg;
						long_msg = NULL;
						long_msg_size = 0;
						ZeroMemory(long_msg_mask,sizeof(long_msg_mask));
					}
				}
			}
        }
        break;
    case CMD_MESSAGE_RECEIVED:
        if(size >= 1+sizeof(u32))
        {
			u32 msg_num = *(u32*)&packet[1];
            MESSAGE *ptr, *prev_ptr;
            for (prev_ptr = ptr = msg; ptr; ptr = ptr->next)
            {
                if (ptr->msg_num == msg_num)
                {
                    if (prev_ptr == ptr) // very first item
                    {
                        msg = ptr->next;
                    }
                    else
                    {
                        prev_ptr->next = ptr->next;
                    }
					delete[] ptr->text;
					PostMessage(parent,WM_MESSAGE_DELIVERED,ptr->min_pos,ptr->max_pos);
                    delete ptr;
                    break;
                }
            }
        }
        break;
    case CMD_HANG_UP:
		switch (state)
		{
		case CTRL_ST_ABORT:
			pack.buffer = &CMD_HANG_DOWN;
			pack.buf_len = sizeof(CMD_HANG_DOWN);
			if (net) net->send_packet(&pack);
			break;
		case CTRL_ST_DIALING:
			state = CTRL_ST_IN_VOICE;
			call_duration = 0;
			buf_overflow_cnt = 0;
			cp_get_time(&call_started);
		case CTRL_ST_IN_VOICE:
			pack.buffer = &CMD_IN_VOICE;
			pack.buf_len = sizeof(CMD_IN_VOICE);
			if (net) net->send_packet(&pack);
			break;
		case CTRL_ST_IDLE:
			state = CTRL_ST_CONNECTED;
		default:
			PostMessage(parent,WM_RING,0,0);
			break;
		}
        break;
    case CMD_HANG_DOWN:
        state = CTRL_ST_IDLE;
		cp_time_t cur_time;
		cp_get_time(&cur_time);
		call_duration = (int)((cur_time-call_started)/10000000);
		PostMessage(parent,WM_CALL_FINISHED,0,0);
        break;
    case CMD_IN_VOICE:
        if (state == CTRL_ST_DIALING)
		{
            state = CTRL_ST_IN_VOICE;
			buf_overflow_cnt = 0;
			call_duration = 0;
			cp_get_time(&call_started);
			voice_buf_filled = false;
			PostMessage(parent,WM_UPDATECHAT,0,0);
		}
        else if (state < CTRL_ST_IN_VOICE)
		{
			pack.buffer = &CMD_HANG_DOWN;
			pack.buf_len = sizeof(CMD_HANG_DOWN);
			if (net) net->send_packet(&pack);
		}
        break;
    case CMD_VOICE:
		voice_pack_num++;
        if (state == CTRL_ST_IN_VOICE)
        {
			int wr_inx;
			EnterCriticalSection(&cs);
			wr_inx = rec_wr_inx + 1;
			if (wr_inx >= CONTR_MAX_VOICE_BUF) wr_inx -= CONTR_MAX_VOICE_BUF;
			if (wr_inx != rec_rd_inx)
			{
				rec_size[rec_wr_inx] = size-1;
				if (rec_size[rec_wr_inx] < 1024)
				{
					memcpy(rec_data[rec_wr_inx],&packet[1],rec_size[rec_wr_inx]);
					rec_wr_inx = wr_inx;
				}
				else
				{
					rec_size[rec_wr_inx] = 0;
				}
			}
			wr_inx = rec_wr_inx - rec_rd_inx;
			if (wr_inx < 0) wr_inx += CONTR_MAX_VOICE_BUF;
			if (wr_inx >= rec_data_buf_limit)
			{
				voice_buf_filled = true;
			}
			if (wr_inx >= 8) buf_overflow_cnt++;
			else buf_overflow_cnt = 0;
			if (buf_overflow_cnt > 500)
			{
				buf_overflow_cnt = 0;
				rec_wr_inx -= 8;
				if (rec_wr_inx < 0) rec_wr_inx += CONTR_MAX_VOICE_BUF;
				if (rec_data_buf_limit > 8 && rec_data_buf_limit > wr_inx) rec_data_buf_limit -= 8;
			}
			LeaveCriticalSection(&cs);
        }
        break;
    case CMD_TYPING:
		PostMessage(parent,WM_SHOW_TYPING,0,0);
        break;
	case CMD_REJECT:
		state = CTRL_ST_REJECTED;
		PostMessage(parent,WM_UPDATECHAT,0,0);
		break;
	}
}

void CONTR_PROT::accept(void)
{
    state = CTRL_ST_IN_VOICE;
	call_duration = 0;
	buf_overflow_cnt = 0;
	cp_get_time(&call_started);
	voice_buf_filled = false;
    voice_cnt = 0;
	PostMessage(parent,WM_UPDATECHAT,0,0);
}

void CONTR_PROT::reject(void)
{
	NETWORK* net = getNet();
 	PACKET_FOR_SEND pack = {0};
	pack.protocol = NETWORK_PROT_CONTROL;
	pack.nd = &nd;

    voice_cnt = 0;
    state = CTRL_ST_ABORT;
	pack.buffer = &CMD_HANG_DOWN;
	pack.buf_len = sizeof(CMD_HANG_DOWN);
	if (net) net->send_packet(&pack);
	PostMessage(parent,WM_UPDATECHAT,0,0);
}

void CONTR_PROT::dial(void)
{
    if (state < CTRL_ST_DIALING)
	{
		NETWORK* net = getNet();
 		PACKET_FOR_SEND pack = {0};
		pack.protocol = NETWORK_PROT_CONTROL;
		pack.nd = &nd;

	    voice_cnt = 0;
		state = CTRL_ST_DIALING;
		pack.buffer = &CMD_HANG_UP;
		pack.buf_len = sizeof(CMD_HANG_UP);
		if (net) net->send_packet(&pack);
		PostMessage(parent,WM_UPDATECHAT,0,0);
	}
}

void CONTR_PROT::hang_down(void)
{
    if (state > CTRL_ST_CONNECTED)
	{
		NETWORK* net = getNet();
 		PACKET_FOR_SEND pack = {0};
		pack.protocol = NETWORK_PROT_CONTROL;
		pack.nd = &nd;

		state = CTRL_ST_IDLE;
		pack.buffer = &CMD_HANG_DOWN;
		pack.buf_len = sizeof(CMD_HANG_DOWN);
		if (net) net->send_packet(&pack);
		cp_time_t cur_time;
		cp_get_time(&cur_time);
		call_duration = (int)((cur_time-call_started)/10000000);
		PostMessage(parent,WM_CALL_FINISHED,0,0);
	}
}

u8* CONTR_PROT::getVoice(u32* size)
{
	u8* buf = NULL;
	*size = 0;
    if ((state != CTRL_ST_IN_VOICE) || !voice_buf_filled) return NULL;
	EnterCriticalSection(&cs);
	if (rec_rd_inx!=rec_wr_inx) // FIFO contains data
	{
		if (rec_size[rec_rd_inx])
		{
			buf = rec_data[rec_rd_inx];
			*size = rec_size[rec_rd_inx];
			rec_size[rec_rd_inx] = 0;
			rec_rd_inx++;
			if (rec_rd_inx >= CONTR_MAX_VOICE_BUF) rec_rd_inx = 0;
		}
		else
		{
			voice_buf_filled = false; // something goes wrong
		}
	}
	else // FIFO underflow
	{
		voice_buf_filled = false;
		if (rec_data_buf_limit < CONTR_MAX_VOICE_BUF-8)
		{
			rec_data_buf_limit+=8;
		}
	}
	LeaveCriticalSection(&cs);
	return buf;
}

void CONTR_PROT::sendTyping(void)
{
	DWORD ticks = GetTickCount();
	if (ticks - type_ticks > 2000)
	{
		NETWORK* net = getNet();
 		PACKET_FOR_SEND pack = {0};
		pack.protocol = NETWORK_PROT_CONTROL;
		pack.nd = &nd;
		pack.buffer = &CMD_TYPING;
		pack.buf_len = sizeof(CMD_TYPING);
		if (net) net->send_packet(&pack);

		type_ticks = ticks;
	}
}

void CONTR_PROT::getVoiceFifo(int* len, int* mean)
{
	if (state == CTRL_ST_IN_VOICE)
	{
		int diff = rec_wr_inx - rec_rd_inx;
		if (diff < 0) diff += CONTR_MAX_VOICE_BUF;
		*len = diff;
	}
	else
	{
		*len = 0;
	}
	*mean = rec_data_buf_limit;
}